(* arhmm.ml
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 *)

open Misc;;
open Random;;
open Arproc;;
(*-------------------------------------------------------------------------*)
class arhmm =
  fun ?prior ?trans ?states ?(k=1) ?(s=1) ?(r=1) () ->
    let s =
      match prior with
      | Some x -> Array.length x
      | None   -> s
    in
(*-------------------------------------------------------------------------*)
    object (self)
      (* Prior and transition matrix *)

      val mutable p =
	match prior with
	| Some x -> x
	| None   -> Array.make s 0.

      val mutable a =
	match trans with
	| Some x -> x
	| None   -> Array.make_matrix s s 0.

      (* AR processes *)

      val mutable arp = 
	match states with
	| Some x -> x
	| None   -> Array.init s (fun _ -> new arproc ~r:r ())

      (* Accumulators *)

      val mutable p1_acc = Array.make s neg_infinity
      val mutable a_acc  = Array.make_matrix s s neg_infinity

      (* Parameters *)

      val mutable k = k
      val mutable s = s
(*-------------------------------------------------------------------------*)
      (* Public methods *)

      method set_p x = p <- x
      method set_a x = a <- x

      method get_s ()   = s
      method get_arp () = arp
(*-------------------------------------------------------------------------*)
      method rnd tt =
	let v   = Array.make tt 0.
	and s   = Array.make tt 0
	and cnt = ref 0 in

	(* Randomly select one state from the prior *)

	let s0 = randgen p in

	s.(0) <- s0;
	v.(0) <- arp.(s0)#rnd v 0;

	(* Other time steps *)

	for t=1 to tt-1 do
	  cnt := (!cnt+k-1) mod k;

	  let stm = s.(t-1) in
	  let st  =
	    if !cnt = 0 then
	      randgen a.(stm)
	    else
	      stm
	  in

	  s.(t) <- st;
	  v.(t) <- arp.(st)#rnd v t
	done;

	(v, s)
(*-------------------------------------------------------------------------*)
      method fwd ?(adapt=true) v =
	let tt = Array.length v in

	let cnt = ref 0
	and llh = ref 0.
	and n   = ref 0
	and rho = Array.make_matrix tt s neg_infinity in

	(* Get the gain contour *)

	let gain =
	  if adapt then
	    self#get_gain v 
	  else
	    let n   = tt / k in
	    let rem = tt mod k in
	    let n   = if rem <> 0 then n+1 else n in

	    let covs = Array.map (fun x -> x#get_cov ()) arp in
	    Array.make n covs
	in
	
	(* First time step *)

	for s0=0 to s-1 do
	  let p1  = log p.(s0)
	  and p2  = arp.(s0)#logpdfcov v 0 gain.(!n).(s0) in
	  let tmp = p1 +. p2 in

	  rho.(0).(s0) <- tmp
	done;

	llh := !llh +. Array.fold_left log_add neg_infinity rho.(0);

	self#normalise rho 0;

	(* Remaining time steps *)

	for t=1 to tt-1 do
	  cnt := (!cnt+k-1) mod k;

	  if !cnt = 0 then n := !n + 1;

	  for st=0 to s-1 do
	    let p1 = arp.(st)#logpdfcov v t gain.(!n).(st) in

	    if !cnt = 0 then
	      for stm=0 to s-1 do
		let p2 = rho.(t-1).(stm)
		and p3 = log a.(stm).(st) in
		
		let tmp1 = p1 +. p2 +. p3
		and tmp2 = rho.(t).(st) in
		let tmp  = log_add tmp1 tmp2 in
		
		rho.(t).(st) <- tmp
	      done
	    else
	      let p2  = rho.(t-1).(st) in
	      let tmp = p1 +. p2 in
	      
	      rho.(t).(st) <- tmp
	  done;
	  
	  llh := !llh +. Array.fold_left log_add neg_infinity rho.(t);
	  
	  self#normalise rho t
	done;

	(rho, !llh /. float_of_int tt, gain)
(*-------------------------------------------------------------------------*)
      method bwd rho =
	let tt = Array.length rho in
	
	let n   = 1 + tt / k in
	let cnt = ref ((n*k - tt + 1) mod k) in

	let post = Array.make_matrix tt s neg_infinity in

	(* Last time step; the last state MUST be s-1 *)

	post.(tt-1).(s-1) <- 0.;

	(* Remaining time steps *)

	for t=tt-2 downto 0 do
	  if !cnt = 0 then
	    let a_bwd = Array.make_matrix s s neg_infinity in

	    for stp=0 to s-1 do
	      for st=0 to s-1 do
		let p1 = log a.(st).(stp)
		and p2 = rho.(t).(st) in

		a_bwd.(stp).(st) <- p1 +. p2
	      done;

	      let sum = Array.fold_left log_add neg_infinity a_bwd.(stp) in
	      
	      if sum <> neg_infinity then
		for st=0 to s-1 do
		  a_bwd.(stp).(st) <- a_bwd.(stp).(st) -. sum
		done
	    done;

	    for st=0 to s-1 do
	      for stp=0 to s-1 do
		let p1 = post.(t).(st)
		and p2 = a_bwd.(stp).(st)
		and p3 = post.(t+1).(stp) in

		post.(t).(st) <- log_add p1 (p2 +. p3)
	      done
	    done

	  else
	    for st=0 to s-1 do
	      post.(t).(st) <- post.(t+1).(st)
	    done;
	    
	  cnt := (!cnt+1) mod k
	done;
	
	(* Compute lambda messages from rho and post *)

	let lambda = Array.make_matrix tt s neg_infinity in

	for t=0 to tt-1 do
	  for st=0 to s-1 do
	    let p1 = rho.(t).(st) in

	    if p1 = neg_infinity then
	      lambda.(t).(st) <- neg_infinity
	    else
	      lambda.(t).(st) <- post.(t).(st) -. p1
	  done
	done;

	(lambda, post)
(*-------------------------------------------------------------------------*)
      method decode ?(adapt=true) v =
	let tt = Array.length v in

	let swap x =
	  let tmp = x.(0) in
	  x.(0) <- x.(1);
	  x.(1) <- tmp
	in

	let cnt   = ref 0
	and n     = ref 0
	and rho   = Array.make_matrix 2 s neg_infinity
	and trace = Array.make_matrix tt s (-1) in
	
	(* Get the gain contour *)

	let gain =
	  if adapt then
	    self#get_gain v 
	  else
	    let n   = tt / k in
	    let rem = tt mod k in
	    let n   = if rem <> 0 then n+1 else n in

	    let covs = Array.map (fun x -> x#get_cov ()) arp in
	    Array.make n covs
	in

	(* First time step *)
	
	for s0=0 to s-1 do
	  let p1  = log p.(s0) 
	  and p2  = arp.(s0)#logpdfcov v 0 gain.(!n).(s0) in
	  let tmp = p1 +. p2 in

	  rho.(0).(s0) <- tmp
	done;
	
	self#normalise rho 0;

	(* Remaining time steps *)

	for t=1 to tt-1 do
	  cnt := (!cnt+k-1) mod k;

	  if !cnt = 0 then n := !n + 1;

	  for st=0 to s-1 do
	    let p1 = arp.(st)#logpdfcov v t gain.(!n).(st) in
	    
	    if !cnt = 0 then
	      let max = ref neg_infinity
	      and idx = ref (-1) in
	      
	      for stm=0 to s-1 do
		let p2 = rho.(0).(stm)
		and p3 = log a.(stm).(st) in
		
		let tmp = p1 +. p2 +. p3 in
		
		if tmp > !max then
		  begin
		    max := tmp;
		    idx := stm;
		  end;		
	      done;
	      
	      trace.(t).(st) <- !idx;
	      rho.(1).(st)   <- !max
	    else
	      let p2 = rho.(0).(st) in
	      trace.(t).(st) <- st;
	      rho.(1).(st)   <- p1 +. p2
	  done;

	  self#normalise rho 1;

	  swap rho;
	done;

	(* Backtrace *)

	let s_seq = Array.make tt (-1) in
	
	s_seq.(tt-1) <- s-1;

	for t=tt-2 downto 0 do
	  let stp = s_seq.(t+1) in
	  s_seq.(t) <- trace.(t+1).(stp);
	done;

	s_seq
(*-------------------------------------------------------------------------*)
      method reset () =
	for i=0 to s-1 do
	  arp.(i)#reset ();
	  p1_acc.(i) <- neg_infinity;

	  for j=0 to s-1 do
	    a_acc.(i).(j) <- neg_infinity
	  done
	done
(*-------------------------------------------------------------------------*)
      method acc ?(adapt=true) v =
	let tt  = Array.length v
	and cnt = ref 0
	and n   = ref 0 in

	(* Fwd/Bwd *)

	let (rho, llh, gain) = self#fwd ~adapt:adapt v in
	let (lambda, post)   = self#bwd rho in
	  
	(* Prior *)

	for s0=0 to s-1 do
	  p1_acc.(s0) <- log_add p1_acc.(s0) post.(0).(s0)
	done;
	
	(* Transition matrix *)

	for t=1 to tt-1 do
	  cnt := (!cnt+k-1) mod k;

	  if !cnt = 0 then
	    let p_stmstgv1T = Array.make_matrix s s neg_infinity in

	    n := !n + 1;

	    for st=0 to s-1 do
	      let p_vt = arp.(st)#logpdfcov v t gain.(!n).(st)
	      and p3   = lambda.(t).(st) in

	      for stm=0 to s-1 do
		let p1  = rho.(t-1).(stm)
		and p2  = log a.(stm).(st) in

		p_stmstgv1T.(stm).(st) <- p1 +. p2 +. p3 +. p_vt
	      done
	    done;

	    (* Normalise *)

	    let sum = Array.fold_left 
		(fun s x -> 
		  log_add s (Array.fold_left log_add neg_infinity x))
		neg_infinity
		p_stmstgv1T
	    in

	    for i=0 to s-1 do
	      for j=0 to s-1 do
		let tmp = p_stmstgv1T.(i).(j) -. sum in
		a_acc.(i).(j) <- log_add a_acc.(i).(j) tmp
	      done
	    done
	done;
	
	(* AR coefficients and covariance *)

	let n = ref 0 in

	cnt := 0;

	for t=0 to tt-1 do
	  for st=0 to s-1 do
	    let pt = exp post.(t).(st) in

	    if adapt then
	      arp.(st)#acc ~prior:pt ~sc:(1. /. gain.(!n).(st)) v t
	    else
	      arp.(st)#acc ~prior:pt v t
	  done;

	  cnt := (!cnt+k-1) mod k;

	  if !cnt = 0 then n := !n + 1;
	done;

	llh
(*-------------------------------------------------------------------------*)
      method update () =
	(* Prior *)

	let sum = Array.fold_left log_add neg_infinity p1_acc in
	
	for i=0 to s-1 do
	  p.(i) <- exp (p1_acc.(i) -. sum)
	done;

	(* Transition matrix *)

	for st=0 to s-1 do
	  let sum = Array.fold_left log_add neg_infinity a_acc.(st) in
	  
	  if sum <> neg_infinity then
	    for stp=0 to s-1 do
	      a.(st).(stp) <- exp (a_acc.(st).(stp) -. sum)
	    done
	  else
	    for stp=0 to s-1 do
	      a.(st).(stp) <- exp a_acc.(st).(stp)
	    done
	done;

	(* AR coefficients and covariance *)

	for st=0 to s-1 do
	  arp.(st)#update ()
	done
(*-------------------------------------------------------------------------*)
      method save output =
	(* Integer parameters *)

	Printf.fprintf output "%d %d 1\n" k s;

	(* Prior and transition matrix *)

	for i=0 to s-1 do
	  Printf.fprintf output "%e\n" p.(i)
	done;

	for i=0 to s-1 do
	  for j=0 to s-1 do
	    Printf.fprintf output "%e " a.(i).(j)
	  done;

	  Printf.fprintf output "\n"
	done;

	(* Accumulators *)

	for i=0 to s-1 do
	  Printf.fprintf output "%e\n" p1_acc.(i)
	done;

	for i=0 to s-1 do
	  for j=0 to s-1 do
	    Printf.fprintf output "%e " a_acc.(i).(j)
	  done;

	  Printf.fprintf output "\n"
	done;

	(* AR processes *)

	for i=0 to s-1 do
	  arp.(i)#save output
	done
(*-------------------------------------------------------------------------*)
      method load input =
	(* Integer parameters *)

	let prms =
	  Array.map (fun x -> int_of_string x) (input_line_split input)
	in

	k <- prms.(0);
	s <- prms.(1);

	(* Prior and transition matrix *)

	p <- Array.init s (fun _ -> float_of_string (input_line input));
	a <- Array.init s (fun _ -> 
	  Array.map (fun x -> float_of_string x) (input_line_split input));

	(* Accumulators *)

	p1_acc <- Array.init s (fun _ -> 
	  float_of_string (input_line input));
	  
	a_acc <- Array.init s (fun _ ->
	  Array.map (fun x -> float_of_string x) (input_line_split input));

	(* AR processes *)

	arp <- Array.init s (fun _ ->
	  let tmp = new arproc () in
	  tmp#load input;
	  tmp)
(*-------------------------------------------------------------------------*)
      (* Private methods *)

      method private normalise x r =
	let sum = Array.fold_left log_add neg_infinity x.(r) in
	
	if sum <> neg_infinity then
	  let len = Array.length x.(r) in
	  
	  for i=0 to len-1 do
	    x.(r).(i) <- x.(r).(i) -. sum
	  done
(*-------------------------------------------------------------------------*)
      method private get_gain v =
	let tt  = Array.length v in
	let n   = tt / k in
	let rem = tt mod k in
	let n   = if rem <> 0 then n+1 else n in

	let gain = Array.make_matrix n s 0. in

	for i=0 to n-1 do
	  for st=0 to s-1 do
	    let g = ref 0. in
	    let c = ref 0.  in
	    
	    let b = i*k in
	    let e = b + (if i = n-1 && rem <> 0 then rem-1 else k-1) in	  
	    
	    for t=b to e do
	      let vt = v.(t) in
	      let m  = arp.(st)#get_mean v t in
	      let dv = vt -. m in
	      
	      g := !g +. dv *. dv;
	      c := !c +. 1.;
	    done;

	    if !g = 0. then
	      gain.(i).(st) <- 1e-15
	    else
	      gain.(i).(st) <- sqrt (!g /. !c);
	    
	  done;
	done;

	gain
(*-------------------------------------------------------------------------*)
    end
