(* arhmm_init.ml
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 *)

open Misc;;
open Arhmm;;
open Arproc;;
open Mydataset;;
open Progressbar;;
(*-------------------------------------------------------------------------*)
let nargs = 2;;
let usage_msg = "Usage: "^Sys.argv.(0)^"\n"^
  "\t[-s <number of states>]\n"^
  "\t[-r <number of AR coefficients]\n"^
  "\t[-k <state duration>]\n"^
  "\t[-f <scale factor>]\n"^
  "\t[-g]\n"^
  "\t<utterances list file>\n"^
  "\t<model file>\n";;

let pbar = new progressbar ();;
(*-------------------------------------------------------------------------*)
let _ =
  (* Handle cmd line arguments *)
  
  let s     = ref 10
  and r     = ref 10
  and k     = ref 140
  and f     = ref 1. 
  and adapt = ref true in

  let arglist =
    [("-s", Arg.Int (fun x -> s := x), 
      "Number of states (10)") ;

     ("-r", Arg.Int (fun x -> r := x), 
      "Number of AR coefficients (10)") ;

     ("-k", Arg.Int (fun x -> k := x), 
      "State duration (140)") ;

     ("-f", Arg.Float (fun x -> f := x), 
      "Scaling factor (1)") ;

     ("-g", Arg.Unit (fun () -> adapt := false), 
      "No gain adaptation (false)") ]
  in
  
  let args = ref [] in
  Arg.parse arglist (fun s -> args := s :: !args) usage_msg;
  
  let args = Array.of_list (List.rev !args) in
  if Array.length args <> nargs then
    begin
      Arg.usage arglist usage_msg;
      exit 1
    end;

  let s = !s
  and r = !r
  and k = !k
  and f = !f
  and adapt  = !adapt
  and ofname = args.(1) in

  (* ... *)

  let dset = new mydataset ~scale:f args.(0) in
  let size = dset#size () in

  let arp = Array.init s (fun _ -> new arproc ~r:r ()) in

  pbar#init ~msg:(string_of_int 0) size 80;

  for i=0 to size-1 do
    pbar#update ~msg:(string_of_int 0) i;

    let seq = dset#get i in
    let tt  = Array.length seq in
    let len = tt / s in

    for j=0 to s-1 do
      let tmp = Array.make len 0. in

      for i=0 to len-1 do
	tmp.(i) <- seq.(j*len+i)
      done;

      for k=0 to len-1 do
	arp.(j)#acc tmp k
      done
    done
  done;

  pbar#finish ~msg:(string_of_int 0) 0;

  for i=0 to s-1 do
    arp.(i)#update ();
    arp.(i)#reset ()
  done;

  (* Prior and transition matrix *)

  let p = Array.make s 0.
  and a = Array.make_matrix s s 0. in

  p.(0) <- 1.;

  for i=0 to s-2 do
    a.(i).(i)   <- 0.5;
    a.(i).(i+1) <- 0.5 
  done;

  a.(s-1).(s-1) <- 1.;

  (* Create an ARHMM *)

  let hmm = new arhmm ~prior:p ~trans:a ~states:arp ~k:k () in

  (* Save the model *)
    
  let output = open_out ofname in
  hmm#save output;
  close_out output;

  (* Realign *)

  hmm#reset ();

  for k=0 to 2 do
    pbar#init ~msg:(string_of_int (k+1)) size 80;

    for i=0 to size-1 do
      pbar#update ~msg:(string_of_int (k+1)) i;

      let seq = dset#get i in
      let tt  = Array.length seq in

      let s_seq = hmm#decode ~adapt:adapt seq in

      for t=0 to tt-1 do
	let st = s_seq.(t) in
	arp.(st)#acc seq t
      done
    done;

    pbar#finish ~msg:(string_of_int (k+1)) 0;

    for i=0 to s-1 do
      arp.(i)#update ();
      arp.(i)#reset ();
    done;

    (* Save the model *)
    
    let output = open_out ofname in
    hmm#save output;
    close_out output;
  done;

  Printf.printf "\n";;
