(* arhmm_train_cov.ml
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 *)

open Misc;;
open Arhmm;;
open Mydataset;;
open Progressbar;;
(*-------------------------------------------------------------------------*)
let nargs = 3;;
let usage_msg = "Usage: "^Sys.argv.(0)^"\n"^
  "\t[-n <max number of iterations>]\n"^
  "\t[-t <threshold>]\n"^
  "\t<input model file>\n"^
  "\t<utterances list file>\n"^
  "\t<output model file>\n";;

let pbar = new progressbar ();;
(*-------------------------------------------------------------------------*)
let _ =
  (* Handle cmd line arguments *)
  
  let n      = ref 1000000
  and thr    = ref 1e-7 in

  let arglist =
    [("-n", Arg.Int   (fun x -> n := x), 
      "Max number of iterations (1e6)");

     ("-t", Arg.Float (fun x -> thr := x), 
      "Threshold (1e-7)") ]
  in
  
  let args = ref [] in
  Arg.parse arglist (fun s -> args := s :: !args) usage_msg;
  
  let args = Array.of_list (List.rev !args) in
  if Array.length args <> nargs then
    begin
      Arg.usage arglist usage_msg;
      exit 1
    end;

  let n      = !n
  and thr    = !thr in

  (* Dataset *)

  let dset = new mydataset args.(1) in
  let size = dset#size () in

  (* Load the ARHMM *)

  let hmm  = new arhmm () in
  let file = open_in args.(0) in
  hmm#load file;
  close_in file;

  let s   = hmm#get_s ()
  and arp = hmm#get_arp () in

  (* ... *)

  let rec loop i old_llh =
    pbar#init ~msg:(string_of_int i) size 80;
    
    let llh     = ref 0.
    and cov_acc = Array.make s 0.
    and p_acc   = Array.make s 0. in

    for j=0 to size-1 do
      let v = dset#get j in
      let (rho, vllh, _)  = hmm#fwd ~adapt:false v in
      let (lambda, post)  = hmm#bwd rho in

      llh := !llh +. vllh;

      (* Accumulate *)

      let tt = Array.length v in

      for t=0 to tt-1 do
	for st=0 to s-1 do
	  let p = exp post.(t).(st)
	  and x = v.(t) -. (arp.(st)#get_mean v t) in

	  cov_acc.(st) <- cov_acc.(st) +. p *. x *. x;
	  p_acc.(st)   <- p_acc.(st) +. p;
	done
      done;

      (* Update the covariance of each state *)

      for st=0 to s-1 do
	let cov = cov_acc.(st) /. p_acc.(st) in
	arp.(st)#set_cov cov
      done;
      
      pbar#update ~msg:(string_of_int i) j
    done;
    
    pbar#finish ~msg:(string_of_int i) 0;
    
    let file = open_out args.(2) in
    hmm#save file;
    close_out file;

    (* ... *)
    
    let llh = !llh /. (float_of_int size) in

    let ratio = (llh -. old_llh) /. abs_float(llh) in
    
    if ratio >= thr && i < n then loop (i+1) llh
  in
  loop 1 neg_infinity;

  Printf.printf "\n";;
