(* arproc.ml
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 *)

open Misc;;
open Matrix;;
open Vector;;
open Gsl_linalg;;
open Gsl_blas;;
(*-------------------------------------------------------------------------*)
class arproc =
    fun ?c ?cov ?(r=1) () ->
      let r = 
	match c with
	| Some x -> Array.length x
	| None   -> r
      in
(*-------------------------------------------------------------------------*)
      object (self)
	(* AR coefficients and covariance matrix *)

	val mutable c =
	  match c with
	    Some x -> x
	  | None   -> Array.make r 0.

	val mutable cov =
	  match cov with
	    Some x -> x
	  | None   -> 0.

	val mutable stored_c   = Array.make r 0.
	val mutable stored_cov = 0.

	(* Accumulators *)
		  
	val mutable p_acc   = 0.
	val mutable b_acc   = Gsl_matrix.create ~init:0. r r
	val mutable d_acc   = Gsl_vector.create ~init:0. r
	val mutable cov_acc = 0.

	val mutable stored_p_acc   = 0.
	val mutable stored_b_acc   = Gsl_matrix.create ~init:0. r r
	val mutable stored_d_acc   = Gsl_vector.create ~init:0. r
	val mutable stored_cov_acc = 0.

	(* Parameters *)

	val mutable r = r
(*-------------------------------------------------------------------------*)
	(* Public methods *)

	method set_cov c = cov <- c

	method get_c ()   = c
	method get_cov () = cov
(*-------------------------------------------------------------------------*)
	method get_mean v t =
	  let m = ref 0. in

	  for i=1 to (min r t) do
	    m := !m -. c.(i-1) *. v.(t-i)
	  done;

	  !m
(*-------------------------------------------------------------------------*)
	method rnd v t =
	  let m = self#get_mean v t in
	  Gauss.rnd m cov
(*-------------------------------------------------------------------------*)
	method logpdf v t =
	  let m = self#get_mean v t in
	  Gauss.log_pdf v.(t) m cov
(*-------------------------------------------------------------------------*)
	method logpdfcov v t c =
	  let m = self#get_mean v t in
	  Gauss.log_pdf v.(t) m c
(*-------------------------------------------------------------------------*)
	method reset () =
	  p_acc   <- 0.;
	  cov_acc <- 0.;

	  Gsl_matrix.set_zero b_acc;
	  Gsl_vector.set_zero d_acc
(*-------------------------------------------------------------------------*)
	method acc ?(prior=1.) ?(sc=1.) v t =
	  let vt  = v.(t)
	  and sc2 = sc *. sc in 

	  (* AR coefficients *)

	  p_acc <- p_acc +. prior;

	  for j=1 to (min r t) do
	    let d_j = Gsl_vector.get d_acc (j-1) in
	    let d_j = d_j +. prior *. vt *. v.(t-j) *. sc2 in

	    Gsl_vector.set d_acc (j-1) d_j;

	    for i=1 to (min r t) do
	      let b_ij = Gsl_matrix.get b_acc (i-1) (j-1) in
	      let b_ij = b_ij +. prior *. v.(t-i) *. v.(t-j) *. sc2 in

	      Gsl_matrix.set b_acc (i-1) (j-1) b_ij
	    done
	  done;

	  (* Covariance matrix *)
	  
	  cov_acc <- cov_acc +. prior *. vt *. vt *. sc2
(*-------------------------------------------------------------------------*)
	method update () =
	  match invert_LU ~protect:true (`M b_acc) with
	  | `M inv_b_acc ->
	      let c_acc = Gsl_vector.create ~init:0. r in
	      symv Upper (-1.) inv_b_acc d_acc 1. c_acc;

	      cov <- cov_acc;

	      for i=0 to r-1 do
		let d_i = Gsl_vector.get d_acc i
		and c_i = Gsl_vector.get c_acc i in

		c.(i) <- c_i; 
		cov   <- cov +. 2. *. d_i *. c_i;

		for j=0 to r-1 do
		  let b_ij = Gsl_matrix.get b_acc i j
		  and c_j  = Gsl_vector.get c_acc j in

		  cov <- cov +. c_i *. b_ij *. c_j
		done
	      done;

	      cov <- cov /. p_acc;

	  | _ ->
	      Printf.fprintf stderr "arproc#update: not defined\n";
	      exit 1
(*-------------------------------------------------------------------------*)
	method save output =
	  (* Parameter *)

	  Printf.fprintf output "%d 1\n" r;

	  (* AR coefficients and covariance matrix *)

	  Array.iter (fun x -> Printf.fprintf output "%e\n" x) c;	  
	  Printf.fprintf output "%e\n" cov;

	  (* Accumulators *)

	  Printf.fprintf output "%e\n" p_acc;

	  save_mat b_acc output;
	  save_vec d_acc output;
	  
	  Printf.fprintf output "%e\n" cov_acc
(*-------------------------------------------------------------------------*)
	method load input =
	  (* Parameters *) 

	  let prms = input_line_split input in

	  r <- int_of_string prms.(0);

	  (* AR coefficients and covariance matrix *)

	  c   <- Array.init r (fun _ -> float_of_string (input_line input));
	  cov <- float_of_string (input_line input);

	  (* Accumulators *)

	  p_acc <- float_of_string (input_line input);

	  b_acc <- Gsl_matrix.create r r;
	  load_mat b_acc input;

	  d_acc <- Gsl_vector.create r;
	  load_vec d_acc input;
	  
	  cov_acc <- float_of_string (input_line input)
(*-------------------------------------------------------------------------*)
	method store () =
	  Array.blit c 0 stored_c 0 r;

	  Gsl_matrix.memcpy b_acc stored_b_acc;
	  Gsl_vector.memcpy d_acc stored_d_acc;

	  stored_cov     <- cov;
	  stored_p_acc   <- p_acc;
	  stored_cov_acc <- cov_acc
(*-------------------------------------------------------------------------*)
	method merge () =
	  for i=0 to r-1 do
	    c.(i) <- c.(i) +. stored_c.(i);
	  done;

	  Gsl_matrix.add b_acc stored_b_acc;
	  Gsl_vector.add d_acc stored_d_acc;

	  cov     <- stored_cov +. cov;
	  p_acc   <- stored_p_acc +. p_acc;
	  cov_acc <- stored_cov_acc +. cov_acc
(*-------------------------------------------------------------------------*)
      end	  
